/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.metamodel.processing;

import java.io.File;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Formatter;

import javax.annotation.Generated;
import javax.annotation.processing.Filer;
import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.TypeElement;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.PrimitiveType;
import javax.lang.model.type.TypeKind;
import javax.lang.model.type.TypeMirror;
import javax.lang.model.util.ElementFilter;
import javax.lang.model.util.Elements;
import javax.tools.FileObject;
import javax.tools.JavaFileObject;
import javax.tools.StandardLocation;

/**
 * {@code MetamodelSource} represents a Acid House's entity metamodel Java 
 * source file.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class MetamodelSource {

	private static final String PACKAGE_TEMPLATE = "package %s;\n\n";

	private static final String IMPORT_TEMPLATE = "import "
			+ Generated.class.getName() + ";\n"
			+ "import org.eiichiro.acidhouse.metamodel.ComparableProperty;\n"
			+ "import org.eiichiro.acidhouse.metamodel.Metamodel;\n\n";

	private static final String CLASS_TEMPLATE = "@Generated(\""
			+ MetamodelGenerator.class.getName() + "\")\n"
			+ "public class %s extends Metamodel<%s> {\n\n" + "%s" + "}\n";

	private static final String CONSTRUCTOR_TEMPLATE = "\tpublic %s() {\n"
			+ "\t\tsuper(%s.class);\n" + "\t}\n\n"
			+ "\tpublic %s(Metamodel<?> parent, String name) {\n"
			+ "\t\tsuper(%s.class, parent, name);\n" + "\t}\n\n";

	private static final String COMPARABLE_PROPERTY_TEMPLATE 
			= "\tpublic ComparableProperty<%s, %s> %s"
			+ " = new ComparableProperty<%s, %s>(this, %s.class, \"%s\");\n\n";

	private static final String METAMODEL_TEMPLATE 
			= "\tpublic %s %s = new %s(this, \"%s\");\n\n";

	private final TypeElement element;

	private final ProcessingEnvironment environment;

	private final String packageName;

	private final String entityName;

	private final String metamodelName;

	public MetamodelSource(TypeElement element, ProcessingEnvironment environment) {
		this.element = element;
		this.environment = environment;
		Elements elementUtils = environment.getElementUtils();
		packageName = elementUtils.getPackageOf(element).getQualifiedName().toString();
		entityName = element.getSimpleName().toString();
		metamodelName = toMetamodelName(entityName);
	}

	/**
	 * Save this instance as a Java source file. Java source file is saved as
	 * the name that adding '_' suffix to the entity class name in the same
	 * package as the entity class. If the source file has been already saved,
	 * this method deletes the file and then creates the file newly.
	 * 
	 * @throws Exception If source file saving is failed due to any exception.
	 */
	public void save() throws Exception {
		Filer filer = environment.getFiler();
		FileObject resource = filer.getResource(StandardLocation.SOURCE_OUTPUT,
				packageName, metamodelName + ".java");
		// environment.getMessager().printMessage(Kind.NOTE,
		// resource.getClass().toString());
		File file = new File(resource.toUri());

		if (file.exists()) {
			file.delete();
		}

		JavaFileObject sourceFile = filer.createSourceFile(packageName + "."
				+ metamodelName, element);
		PrintWriter writer = new PrintWriter(sourceFile.openWriter());
		writer.printf(PACKAGE_TEMPLATE, packageName);
		writer.printf(IMPORT_TEMPLATE);
		StringBuilder body = new StringBuilder(new Formatter().format(
				CONSTRUCTOR_TEMPLATE, metamodelName, entityName, metamodelName,
				entityName).toString());
		Elements elementUtils = environment.getElementUtils();
		int i = 0;

		for (VariableElement field : ElementFilter.fieldsIn(
				elementUtils.getAllMembers(element))) {
			if (!isTransient(field)) {
				if (isIdProperty(field)) {
					if (++i > 1) {
						throw new Exception("Entity must have one @Id property");
					}
					
					if (!isIdType(field)) {
						throw new Exception("@Id property must be "
								+ "java.lang.Long or java.lang.String or long");
					}
				}
				
				String name = field.getSimpleName().toString();

				if (isComparableProperty(field)) {
					TypeMirror type = field.asType();
					body.append(new Formatter().format(
							COMPARABLE_PROPERTY_TEMPLATE, entityName,
							box(type), name, entityName, box(type), box(type),
							name).toString());
				} else if (isArray(field)) {
					// Array property is not supported in current version.
					body.append("\t// Array property (" + field
							+ ") is not supported in current version.\n\n");
				} else if (isCollection(field)) {
					// Collection property is not supported in current version.
					body.append("\t// Collection property (" + field
							+ ") is not supported in current version.\n\n");
				} else {
					// Owned child object.
					String metamodel = toMetamodelName(field.asType()
							.toString());
					body.append(new Formatter().format(METAMODEL_TEMPLATE,
							metamodel, name, metamodel, name));
				}
			}
		}
		
		if (i == 0) {
			throw new Exception("Entity must have one @Id property");
		}

		writer.printf(CLASS_TEMPLATE, metamodelName, entityName, body);
		writer.close();
	}

	private String toMetamodelName(String type) {
		// To be configurable?
		return type + "_";
	}

	private boolean isTransient(VariableElement field) {
		for (AnnotationMirror annotationMirror : field.getAnnotationMirrors()) {
			if (annotationMirror.getAnnotationType().toString().equals(
					"org.eiichiro.acidhouse.Transient")) {
				return true;
			}
		}

		return false;
	}

	private static final String[] COMPARABLES = { "boolean",
			"com.google.appengine.api.datastore.Key",
			"com.google.appengine.api.datastore.Link",
			"com.google.appengine.api.datastore.ShortBlob",
			"com.google.appengine.api.users.User", "double", "float", "int",
			"java.lang.Boolean", "java.lang.Double", "java.lang.Float",
			"java.lang.Integer", "java.lang.Long", "java.lang.Short",
			"java.lang.String", "java.util.Date", "long", "short" };

	private boolean isComparableProperty(VariableElement field) {
		String type = field.asType().toString();
		return Arrays.binarySearch(COMPARABLES, type) >= 0;
	}
	
	private boolean isIdProperty(VariableElement field) {
		for (AnnotationMirror mirror : field.getAnnotationMirrors()) {
			if (mirror.getAnnotationType().toString().equals("org.eiichiro.acidhouse.Id")) {
				return true;
			}
		}
		
		return false;
	}

	private static final String[] ID_TYPES = {
		"java.lang.Long", "java.lang.String", "long"
	};
	
	private boolean isIdType(VariableElement field) {
		String type = field.asType().toString();
		return Arrays.binarySearch(ID_TYPES, type) >= 0;
	}
	
	private static final String[] COLLECTIONS = { "java.util.ArrayList",
			"java.util.HashSet", "java.util.LinkedHashSet",
			"java.util.LinkedList", "java.util.List", "java.util.Set",
			"java.util.SortedSet", "java.util.Stack", "java.util.TreeSet",
			"java.util.Vector" };
	
	private boolean isArray(VariableElement field) {
		return (field.asType().getKind() == TypeKind.ARRAY);
	}

	private boolean isCollection(VariableElement field) {
		String type = ((TypeElement) environment.getTypeUtils().asElement(
				field.asType())).getQualifiedName().toString();
		return Arrays.binarySearch(COLLECTIONS, type) >= 0;
	}

	private String box(TypeMirror type) {
		if (type.getKind().isPrimitive()) {
			return environment.getTypeUtils().boxedClass((PrimitiveType) type).toString();
		}

		return type.toString();
	}

}
